<?php

namespace Drupal\mcapi;
use Drupal\Core\Render\Markup;

/**
 * Provides upcasting for the transaction/{xid} path
 *
 * If {serial} is 0, the transaction will be pulled from the private tempstore
 * for the current user.
 */
class CurrencyFormatter {

  /**
   * The currency format, retrieved from config. An array with keys
   * f1 - Formatting to put before the value
   * main - a zero, optionally followed optionally by a decimal point and any number of nines.
   * f2 - Formatting string to put between the units and subdivisions.
   * subdivision - the max number of subunits followed optionally by a slash and the number to divide the unit into e.g. 59/4
   * f3 - Formatting to put after the value.
   * @var array
   */

  /**
   * The 'main' part of the formatting string
   * @var string
   */
  private ?string $prefix;
  private ?string $separator;
  private ?string $suffix;
  private bool $fractions;

  /**
   * The 'subdivision' part of the formatting string
   * @var string
   * @example 99, 59, 59/4,
   */
  private ?int $denominator = 0;
  private ?int $maxDivisions = 0;
  /**
   * How many parts to divide the unit into, for UI purposes.
   * @var int|null
   */
  private ?int $options = NULL;

  /**
   * Constructs a new CurrencyFormatter.
   */
  public function __construct($config_factory) {
    $currency = $config_factory->get('mcapi.settings')->get('currency');
    $this->fractions = $currency['fractions'] ?? FALSE;
    $format = $currency['format'];
    $this->prefix = $format['f1'];
    $this->separator = $format['f2'] ?? '';
    $this->suffix = $format['f3'] ?? '';

    if (is_string($format['subdivision']) and strpos($format['subdivision'], '/')) {
      [$this->maxDivisions, $this->options] = explode('/', $format['subdivision']);
    }
    elseif ($dotpos = strpos($format['main'], '.')) {
      $this->maxDivisions = substr($format['main'], $dotpos+1);
      $this->separator = '.';
    }
    elseif (is_numeric($format['subdivision'])) {
      $this->maxDivisions = $format['subdivision'];
    }
    $this->maxDivisions = intval($this->maxDivisions);
    $this->denominator =  $this->maxDivisions + 1;
  }

  /**
   * Format a currency value.
   * @param int $raw_num
   * @return string (or even Markup?)
   */
  public function format(int|float $raw_num) : Markup {
    [$main, $raw_units] = self::formatParts(abs($raw_num));
    // It's all about rendering the subdivision.
    if ($this->denominator and $raw_units) {
      if ($this->fractions) { // Specific $raw_units allowed
        $frac = self::decToFraction($raw_units / $this->denominator);
        // Frac is a unicode string denoting a fraction
        if ($frac === '0') { // Rounded down
          $subdivision = '';
        }
        elseif ($frac === '1') { // rounded up
          $main++;
          $subdivision = '';
        }
        elseif ($frac) {
          $subdivision = $frac;
        }
      }
      $separator = '';
    }
    if (!isset($subdivision) and $this->denominator) {// just zeros after decimal point
      if (!$this->fractions) {
        // Show raw units after the separator
        $subdivision = str_pad(strval($raw_units), strlen($this->maxDivisions), '0', STR_PAD_LEFT);
      }
      else {
        $subdivision = '';
      }
      $separator = $subdivision ? $this->separator : '';
    }

    $minus = ($raw_num < 0) ? '-' : '';
    $formatted = $minus . $this->prefix . $main . $separator . $subdivision . $this->suffix;
    return Markup::Create(html_entity_decode($formatted, ENT_QUOTES, 'UTF-8'));
  }

  /**
   * Format the value as a machine-readable number
   *
   * @param type $raw_num
   * @return float
   */
  public function formatAsFloat(int|float $raw_num) : float {
    [$main, $raw_units] = self::formatParts(abs($raw_num));
    $float = $main;
    if (!empty($this->denominator) and isset($raw_units)) {
      $float += $raw_units/$this->denominator;
    }
    if ($raw_num < 0) {
      // always render it as a positive number.
      $float *= -1;
    }
    return round($float, 2);
  }

  /**
   * Convert the raw value into numbers for a main widget and subdivision widget.
   *
   * @param NULL|int|string $raw_num
   * @return array
   */
  public function formatParts(NULL|int|float $raw_num) : array {
    if (!is_numeric($raw_num)) {
      $main = 0;
      $raw_units = NULL;
    }
    else{
      if ($this->denominator) {
        $main = intdiv(abs($raw_num), $this->denominator);
        $raw_units = abs($raw_num) % $this->denominator;
      }
      else {
        $main = $raw_num;
      }
    }
    return [$main, $raw_units];
  }


  /**
   * Generate the form element from a raw value.
   *
   * @param $default_val
   *   Whatever was passed to the form element #default_value, null, string or number.
   * @return array
   * @todo typehint $minus when all field settings are updated.
   */
  public function generateFormElements($default_val, int $placeholder_val, $minus = FALSE) {
    if (is_numeric($default_val)) {
     [$main, $subdivision] = self::formatParts(abs($default_val));
    }
    else {
      $main = '';
      $subdivision = 0;
    }
    $elements['main'] =  [
      '#type' => 'number',
      '#default_value' => strval($main),
      '#size' => 6, // makes no difference because of other css.
      '#weight' => 1,
      '#pattern' => '[0-9]+',
      '#maxlength' => '10',
      '#attributes' => ['style' => 'width:5em'],
      '#field_prefix' => ($minus ? '-' : '') .$this->prefix,
      '#min' => 0
    ];

    if ($this->separator == '.') {
      if ($main) {
        $elements['main']['#default_value'] .= '.' . str_pad($subdivision, strlen($this->maxDivisions), '0', STR_PAD_LEFT);
      }
      $elements['main']['#step'] = pow(10, -strlen($this->maxDivisions));
      $elements['main']['#pattern'] .= '(\.[0-9]{'.strlen($this->maxDivisions).'})?';
    }
    elseif ($this->denominator) {
      $default_div = intval($default_val) % $this->denominator;
      $size = strlen($this->maxDivisions);
      $elements['main']['#suffix'] = $this->separator;
      // If there is a slash in the format, then prepare a dropdown.
      if ($this->options) {
        $options = [];
        $step = $this->denominator / $this->options;
        for ($j = 0; $j < $this->options; $j++) {
          $v = $j * $step;
          $options[intval($v)] = $v;
        }
        // If a preset value isn't in the $options ignore the options and use
        // the numeric sub-widget.
        if (array_key_exists(intval($default_div), $options)) {
          $elements['subdivision'] = [
            '#type' => 'select',
            '#options' => $options,
            '#weight' => 3
          ];
        }
        else {
          $elements['subdivision'] = [
            '#type' => 'number',
            '#min' => 0,
            '#step' => $step,
            '#size' => $size,
            '#maxlength' => $size,
            '#attributes' => ['style' => 'width:' . ($size + 3) . 'em;'],
            '#max' => $this->maxDivisions, //Which is already all the nines
          ];
        }
      }
      else {// Just enter the number of sub units
        $elements['subdivision'] = [
          '#type' => 'number',
          '#min' => 0,
          '#step' => 1,
          // #size has no effect in opera.
          '#size' => $size + 2, // Spinner takes about 2 chars
          '#maxlength' => $size,
          '#max' => $this->maxDivisions, // E.G. 59 minutes
        ];
      }
      $elements['subdivision'] += [
        '#suffix' => $this->suffix,
        '#default_value' => $default_div,
        '#attributes' => ['class' => ['worth-subdivision']],
        '#theme_wrappers' => [],
        '#placeholder' => str_pad('',$size, '0'),
        '#weight' => 3
      ];
    }
    // Placeholder
    extract($this->formatParts($placeholder_val));
   [$main, $subdivision] = self::formatParts(abs($placeholder_val));
    $elements['main']['#placeholder'] = $main;
    if ($this->separator == '.') {
      $elements['main']['#placeholder'] .= '.' . str_pad($subdivision, strlen($this->maxDivisions), '0', STR_PAD_LEFT);
    }

    return $elements;
  }

  /**
   * Convert the two part of a number (from the form element) into the raw value.
   * @param string $main
   * @param string|null $subdivision
   * @return int
   */
  public function unformat(string $main, ?string $subdivision = '') : int {
    $main = $main ? $main + 0 : 0; //Make this numeric if it is null.
    if (empty($this->denominator)) {
      $raw_num = $main;
    }
    else {
      $raw_units = is_numeric($subdivision) ? $subdivision : 0;
      $raw_num = ($main * $this->denominator) + $raw_units;
    }
    return intval($raw_num);
  }

  /**
   * Express a float < 1 as half, thirds, quarters or fifths.
   * @param float $float
   * @return string
   */
  private static function decToFraction($float) : string {
    if ($float >= 1) {
      throw new \Exception("Only convert numbers less than one into options: $float");
    }
    $numerator = floor($float * 60);
    // HTML and this function support all the denominators up to 5, w
    // Wich means we calculate using 60ths
    $options = [
      0 => '0',
      7 => '&#8533;', #1/5
      13 => '&#188;',  #1/4
      18 => '&#8531;', #1/3
      22 => '&#8534;', #2/5
      26 => '&#189;',  #1/2
      34 => '&#8535;', #3/5
      38 => '&#8532;', #2/3
      42 => '&#190;',  #3/4
      50 => '&#8536;', #4/5
      // anything else rounded up
    ];
    $last = 0;
    foreach (array_keys($options) as $sixtieths) {
      if ($numerator >= $sixtieths) {
        $last = $sixtieths;
        continue;
      }
      return $options[$last];
    }
  }

  /**
   * @param int $raw_num
   *   $num formatted as float.
   * @return string[]
   *   The formatted axis points, keyed by float value.
   */
  function axisMarkers(int $raw_num) {
    if ($raw_num == 0) {
      $raw_num = $this->unformat(1);
    }
    $float = $this->formatAsFloat($raw_num);
    $exponent = floor(log10(abs($float)));
    $mantissa = abs($float / pow(10, $exponent));
    // Transform the number into a number between 1 and 10
    if ($mantissa > 7.5) $markers = [5, 10];
    elseif ($mantissa > 6) $markers = [2.5, 5, 7.5];
    elseif ($mantissa > 4.5) $markers = [2, 4, 6];
    elseif ($mantissa > 4) $markers = [1.5, 3, 4.5];
    elseif ($mantissa > 3) $markers = [2, 4];
    elseif ($mantissa > 2) $markers = [1, 2, 3];
    elseif ($mantissa > 1.5) $markers = [1, 2];
    else $markers = [1, 2];

    // Transform the markers back to the right power of 10
    if ($this->denominator) {
      array_walk(
        $markers,
        function(&$n) use ($exponent, $float, $raw_num) {
          $n *= pow(10, $exponent);
          $n *= $this->denominator;
          if ($float < 0) $n *= -1;
          $n = intval($n);
        }
      );
    }
    $markers = [0] + $markers;

    foreach ($markers as $raw_worth) {
      $label = strip_tags((string)$this->format(intval($raw_worth)));
      $result[$label] = $this->formatAsFloat($raw_worth);
    }
    return $result;
  }

  /**
   * Calculate some good axis labels, using the min & max balance extents.
   *
   * @param $raw_min
   * @param $raw_max
   *
   * @return string[]
   *   $min, $mid, $max axis labels keyed by float values.
   */
  function gchartAxes(int $raw_min, int $raw_max) : array {
    $range = $raw_max - $raw_min;
    if ($raw_min >= 0) { // All positive
      $markers = $this->axisMarkers($raw_max);
    }
    elseif ($raw_max <= 0) {// All negative
      $markers = $this->axisMarkers($raw_min);
    }
    else { // Equal axes either side of zero.
      $markers = [];
      $pos_markers = $this->axisMarkers(max(abs($raw_min), abs($raw_max)));
      foreach ($pos_markers as $string => $float) {
        $markers[$string] = $float;
        $markers["-$string"] = -$float;
      }
      unset($markers['-0']);
      asort($markers);
    }
    return $markers;
  }


  function __gchartAxes($min, $max) : array {
    $range = $max - $min;
    if ($min >= 0) { // All positive
      $markers = $this->axisMarkers($max);
    }
    elseif ($max <= 0) {// All negative
      $markers = $this->axisMarkers($min);
    }
    else { // either side of zero.
      $pos = $this->axisMarkers($max);
      $neg= $pos;
      array_walk($neg, function (&$n) {$n = -$n;});
      unset ($neg[0]);
      $markers = array_merge($neg, $pos);

    }
    return $markers;
  }

}
